from rgtracker.common import *
from rgtracker.record import *
from rgtracker.tracker import *
from rgtracker.website import *
from rgtracker.section import *
from rgtracker.page import *
from rgtracker.device import *
from redisgears import executeCommand as execute
import json
import sys
import datetime

import time
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry


def requests_retry_session(
        retries=3,
        backoff_factor=0.3,
        status_forcelist=(500, 502, 504, 404),
        session=None,
):
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session


def enrich_page(record):
    key = record['value']['key']
    article_id = record['value']['article_id']
    api_url = f"http://rtl-curator.back.k8s.rtl/items/urn:newstool:article:{article_id}"
    try:
        response_json = requests_retry_session().get(api_url).json()
        element = datetime.datetime.strptime(response_json.get('display_date'), "%Y-%m-%dT%H:%M:%S%z")
        timestamp = round(datetime.datetime.timestamp(element) * 1000)
        result = {
            'kicker': response_json.get('kicker'),
            'title': response_json.get('title'),
            'display_date': timestamp
        }
        execute('JSON.SET', key, '$.metadata', json.dumps(result))
        # tracker_log(f'{key} - {result} - OK', prefix='Enrich - ')
    except requests.exceptions.RequestException as e:
        # tracker_log(f'{e} for {key}', prefix='Enrich - ', log_level='warning')
        pass


tracker_log(f'Register EnrichPage ...')

desc_json = {
    "name": 'Enrich',
    "version": '99.99.99',
    "desc": f"Enrich Website, Section and Page"
}
# unregister_old_versions(desc_json.get('name'), desc_json.get('version'))
GB("StreamReader", desc=json.dumps(desc_json)). \
    foreach(enrich_page). \
    register(
    prefix='ST:ENRICH:P:::',
    convertToStr=True,
    collect=True,
    onFailedPolicy='abort',
    onFailedRetryInterval=1,
    batch=1,
    duration=0,
    trimStream=False)

tracker_log(f'Register EnrichPage OK')



from rgtracker.record import *
from rgtracker.tracker import *
from rgtracker.common import *
from rgtracker.website import *
from rgtracker.section import *
from rgtracker.page import *
from rgtracker.device import *
from rgtracker.pageviews import *

GB("StreamReader"). \
    aggregate([],
              lambda a, r: a + [r['value']],
              lambda a, r: a + r). \
    run('ST:5MINUTES:W:::UD', trimStream=False, fromId="1663766604999")

# register(
# prefix=f'ST:1MINUTE:{dimension}:::PG',
# convertToStr=True,
# collect=True,
# onFailedPolicy='abort',
# onFailedRetryInterval=1,
# batch=1,
# duration=0,
# trimStream=False)

# run('ST:1MINUTE::::PG', trimStream=False)



from .record import Record

from .tracker import Tracker

from .website import (
    Website,
    load_website
)

from .section import Section

from .page import Page

from .device import Device

__all__ = [
    "Record",
    "Tracker",
    "Website",
    "load_website",
    "Section",
    "Page",
    "Device",
]



import redis
import json

stream_key = 'ST:TRACKER::::'

if __name__ == '__main__':
    # r = redis.Redis(host='localhost', port=6379, db=0)
    redis_dev = redis.Redis(host='10.61.228.218', port=6379, db=0, password='PiPaPo-2021.12')
    redis_local = redis.Redis(host='localhost', port=6379, db=0)
    while True:
        l = redis_dev.xread(count=1, block=5000, streams={stream_key: '$'})[-1][-1][-1][-1]
        # l = { key.decode(): val.decode() for key, val in l.items() }
        print(f"stream length: {redis_dev.xlen(stream_key)}")
        print(f"{redis_dev.xlen(stream_key)} - {type(l)} {l}")

        redis_local.xadd(stream_key, l)

# d = [b'ST:TRACKER::::', [(b'1664453926273-0',
#                           {b'url': b'https://www.rtl.lu/news/panorama/a/1972529.html?didomiConfig.notice.enable=false',
#                            b'uui': b'ipdigl5f9zzzs0q10k39ixfa',
#                            b'ua': b'Mozilla/5.0 (Linux; Android 12; SM-G996B Build/SP1A.210812.016; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/105.0.5195.136 Mobile Safari/537.36',
#                            b'lang': b'de-DE,de;q=0.9,fr-LU;q=0.8,fr;q=0.7,en-DE;q=0.6,en-US;q=0.5,en;q=0.4',
#                            b'dt': b'1664453925996'})]]
# print(type(d[-1][-1][-1]), d[-1][-1][-1])


from redisgears import executeCommand as execute
from redisgears import log
import json
from datetime import datetime, timedelta
from enum import Enum, IntEnum, unique

NAME = "TRACKER"


def tracker_log(msg, prefix=f'{NAME} - ', log_level='notice'):
    msg = prefix + msg
    log(msg, level=log_level)


def tracker_debug(msg):
    tracker_log(msg, log_level='debug')


# from utils import Type, Dimension, RedisNC, Metric, create_key_name, parse_key_name, round_time

@unique
class RedisNC(IntEnum):
    TYPE = 0,
    NAME = 1,
    DIMENSION = 2,
    RECORD_ID = 3,
    TS = 4,
    METRIC = 5


@unique
class Type(Enum):
    STREAM = 'ST'
    HASH = 'H'
    JSON = 'J'
    INDEX = 'I'
    TIMESERIES = 'TS'
    BLOOM = 'B'
    SORTEDSET = 'SS'
    SET = 'S'
    LIST = 'L'
    CHANNEL = 'C'
    CMS = 'CMS'
    HLL = 'HLL'


@unique
class Dimension(Enum):
    WEBSITE = 'W'
    SECTION = 'S'
    PAGE = 'P'
    DEVICE = 'D'
    AUDIO = 'A'
    VIDEO = 'V'
    PODCAST = 'PC'
    METRIC = 'M'


@unique
class Metric(Enum):
    PAGEVIEWS = 'PG'
    DEVICES = 'D'
    UNIQUE_DEVICES = 'UD'


def create_key_name(type, name='', dimension='', record_id='', ts='', metric=''):
    return f'{type}:{name}:{dimension}:{record_id}:{ts}:{metric}'


def parse_key_name(key):
    def check_none(check_value):
        if check_value == '':
            return None
        else:
            return check_value

    key_split = key.split(':')
    type = check_none(key_split[RedisNC.TYPE])
    name = check_none(key_split[RedisNC.NAME])
    dimension = check_none(key_split[RedisNC.DIMENSION])
    record_id = check_none(key_split[RedisNC.RECORD_ID])
    ts = check_none(key_split[RedisNC.TS])
    metric = check_none(key_split[RedisNC.METRIC])

    return {
        'type': type,
        'name': name,
        'dimension': dimension,
        'record_id': record_id,
        'ts': ts,
        'metric': metric
    }


def round_time(dt=None, round_to=60):
    if dt is None: dt = datetime.now()
    seconds = (dt.replace(tzinfo=None) - dt.min).seconds
    rouding = (seconds + round_to / 2) // round_to * round_to
    return dt + timedelta(0, rouding - seconds, -dt.microsecond)


def convert_list_to_dict(x):
    it = iter(x)
    res_dct = dict(zip(it, it))
    return res_dct


def get_ts(ts, ts_arr):
    if len(ts.split('_')) > 1:
        return f'{ts_arr[0].get("ts").split("_")[0]}_{ts_arr[-1].get("ts").split("_")[-1]}'
    else:
        return f'{ts_arr[0].get("ts")}_{ts_arr[-1].get("ts")}'


def get_ts_df(ts, ts_arr):
    if len(ts.split('_')) > 1:
        return f'{ts_arr.iloc[0].split("_")[0]}_{ts_arr.iloc[-1].split("_")[-1]}'
    else:
        return f'{ts_arr.iloc[0]}_{ts_arr.iloc[-1]}'


def registration_arr_to_dict(registration, depth):
    if depth >= 2:
        return registration
    if type(registration) is not list:
        return registration
    d = {}
    for i in range(0, len(registration), 2):
        d[registration[i]] = registration_arr_to_dict(registration[i + 1], depth + 1)
    return d


def compare_versions(v1, v2):
    # None version is less then all version
    if v1 is None:
        return -1
    if v2 is None:
        return 1

    if v1 == "99.99.99":
        return 1
    if v2 == "99.99.99":
        return -1

    v1_major, v1_minor, v1_patch = v1.split(".")
    v2_major, v2_minor, v2_patch = v2.split(".")

    if int(v1_major) > int(v2_major):
        return 1
    elif int(v1_major) < int(v2_major):
        return -1

    if int(v1_minor) > int(v2_minor):
        return 1
    elif int(v1_minor) < int(v2_minor):
        return -1

    if int(v1_patch) > int(v2_patch):
        return 1
    elif int(v1_patch) < int(v2_patch):
        return -1

    return 0


def unregister_old_versions(name, version):
    tracker_log(f"Unregister old version of {name}")
    registrations = execute("RG.DUMPREGISTRATIONS")
    for registration in registrations:
        registration_dict = registration_arr_to_dict(registration, 0)
        desc_str = registration_dict["desc"]
        try:
            desc = json.loads(desc_str)
        except Exception as e:
            continue
        if "name" in desc.keys() and name in desc["name"]:
            tracker_log(
                "Version auto upgrade is not atomic, make sure to use it when is not traffic to the database (otherwise you might lose events).",
                log_level="warning"
            )
            if "version" not in desc.keys():
                execute("RG.UNREGISTER", registration_dict["id"])
                tracker_log(f"Unregister {registration_dict['id']}")
                continue
            v = desc["version"]
            if compare_versions(version, v) > 0:
                execute("RG.UNREGISTER", registration_dict["id"])
                tracker_log(f"Unregistered {registration_dict['id']}")
            else:
                raise Exception(
                    "Found a version which is greater or equals current version, aborting."
                )
    tracker_log("Unregistered old versions.")


def get_maxlen_capped_stream(interval, dimension):
    """
    Save on day of rotated data, approximation.
    Example:
    5MINUTES interval -> (1440 / 5) * number_of_items_in_dimensions
    :param interval:
    :param dimension:
    :return:
    """
    one_day_in_minutes = 1440
    nb_of_websites = 6
    nb_of_sections_by_website = 200
    nb_of_pages_by_section = 3000
    # Python 3.10, but redisgears ... Python version 3.7.2 ...
    # match interval:
    #     case '5MINUTES':
    #         match dimension:
    #             case Dimension.WEBSITE.value:
    #                 return (one_day_in_minutes / 5) * nb_of_websites
    #             case Dimension.SECTION.value:
    #                 return (one_day_in_minutes / 5) * (nb_of_websites * nb_of_sections_by_website)
    #             case Dimension.PAGE.value:
    #                 return (one_day_in_minutes / 5) * (nb_of_websites * nb_of_sections_by_website * nb_of_pages_by_section)
    if interval == '5MINUTES':
        if dimension == Dimension.WEBSITE.value:
            # return int((one_day_in_minutes / 6) * nb_of_websites)
            return (5 * 6) * nb_of_websites
        elif dimension == Dimension.SECTION.value:
            # return int((one_day_in_minutes / 6) * (nb_of_websites * nb_of_sections_by_website))
            return (5 * 6) * nb_of_websites * nb_of_sections_by_website
        elif dimension == Dimension.PAGE.value:
            # return int((one_day_in_minutes / 6) * (nb_of_websites * nb_of_sections_by_website * nb_of_pages_by_section))
            return (5 * 6) * nb_of_websites * nb_of_sections_by_website * nb_of_pages_by_section
    elif interval == '10MINUTES':
        if dimension == Dimension.WEBSITE.value:
            return (6 * 10) * nb_of_websites
        elif dimension == Dimension.SECTION.value:
            return (6 * 10) * nb_of_websites * nb_of_sections_by_website
        elif dimension == Dimension.PAGE.value:
            return (6 * 10) * nb_of_websites * nb_of_sections_by_website * nb_of_pages_by_section
    elif interval == '1HOUR':
        if dimension == Dimension.WEBSITE.value:
            return (1 * 24) * nb_of_websites
        elif dimension == Dimension.SECTION.value:
            return (1 * 24) * nb_of_websites * nb_of_sections_by_website
        elif dimension == Dimension.PAGE.value:
            return (1 * 24) * nb_of_websites * nb_of_sections_by_website * nb_of_pages_by_section
    elif interval == '3HOURS':
        if dimension == Dimension.WEBSITE.value:
            return (3 * 56) * nb_of_websites
        elif dimension == Dimension.SECTION.value:
            return (3 * 56) * nb_of_websites * nb_of_sections_by_website
        elif dimension == Dimension.PAGE.value:
            return (3 * 56) * nb_of_websites * nb_of_sections_by_website * nb_of_pages_by_section


def rotate_agg(x):
    if len(x) > 1:
        z = []
        for s in x:
            z.append(s)
        return z
    else:
        return x


def is_weird(last_visited, last_visited_rounded):
    n = datetime.now()
    now = round_time()
    now_inf = now - timedelta(minutes=0.5)
    now_sup = now + timedelta(minutes=0.5)
    record_rounded_dt = datetime.fromtimestamp(last_visited_rounded / 1000)
    record_dt = datetime.fromtimestamp(int(last_visited) / 1000)

    if now_inf <= record_dt <= now_sup:
        # tracker_log(f'{record_dt} -> {record_rounded_dt} between [{now_inf} {now} {now_sup}] <- {n}')
        return False
    else:
        tracker_log(f'{record_dt} -> {record_rounded_dt} is not between [{now_inf} {now} {now_sup}] <- {n}')
        return True



from rgtracker.record import *
from rgtracker.tracker import *
from rgtracker.common import *
from rgtracker.website import *
from rgtracker.section import *
from rgtracker.page import *
from rgtracker.device import *
from rgtracker.pageviews import *

job = {
    'name': 'W-PG-5to10',
    'version': '99.99.99',
    'input_stream_name': create_key_name(Type.STREAM.value, '5MINUTES', Dimension.WEBSITE.value, '', '', Metric.PAGEVIEWS.value),
    'dimension': Dimension.WEBSITE.value,
    'number_of_rotated_keys': 2, # rotate 2 keys of 5 minutes each
    'write_to_ts': False,
    'timeseries_name': '10MINUTES',
    # 'key_expire_duration_sc': 3620, # keep 6 keys -> merged key expire 60 minutes later
    'key_expire_duration_sc': 7240, # TEST
    'batch_size': 9999999,
    'batch_interval_ms': 600000, # run the job every 10 minutes
    'output_stream_name': create_key_name(Type.STREAM.value, '10MINUTES', Dimension.WEBSITE.value, '', '', Metric.PAGEVIEWS.value)
}

# Todo: unregister_old_versions job
# unregister_old_versions(job.get('name'), job.get('version'))

tracker_log(f'Register {job.get("name")} ...')

desc_json = {
    "name": job.get('name'),
    "version": job.get('version'),
    "desc": f"Rotate Website Pageviews Keys, 5 minutes to 10 minutes buckets"
}

GB("StreamReader", desc=json.dumps(desc_json)). \
    aggregate([],
              lambda a, r: a + [r['value']],
              lambda a, r: a + r). \
    map(lambda records: transform(records, job.get('number_of_rotated_keys'), job.get('dimension'))). \
    foreach(lambda records: load(
        job.get('name'),
        records,
        job.get('dimension'),
        job.get('write_to_ts'),
        job.get('timeseries_name'),
        job.get('key_expire_duration_sc'),
        job.get('input_stream_name'),
        job.get('output_stream_name')
    )). \
    foreach(lambda records: tracker_log(f'ids:{len(records.get("ids"))} merged:{len(records.get("merge"))} reinjected:{len(records.get("reinject"))} - {records}', f'{job.get("name")} - ')). \
    register(
        prefix=job.get('input_stream_name'),
        convertToStr=True,
        collect=True,
        onFailedPolicy='abort',
        onFailedRetryInterval=1,
        batch=job.get('batch_size'),
        duration=job.get('batch_interval_ms'),
        trimStream=False)

tracker_log(f'Register {job.get("name")} OK')



from rgtracker.record import *
from rgtracker.tracker import *
from rgtracker.common import *
from rgtracker.website import *
from rgtracker.section import *
from rgtracker.page import *
from rgtracker.device import *
from rgtracker.pageviews import *
from redisgears import log

# desc_json_s = {"name": 'S'}
# GB("StreamReader", desc=json.dumps(desc_json_s)). \
#     filter(lambda record: record['value']['dimension'] == 'S'). \
#     foreach(lambda record: log(f'sandbox-s-{record}')). \
#     register(
#     prefix='ST:1MINUTE:S:::PG',
#     convertToStr=True,
#     collect=True,
#     onFailedPolicy='abort',
#     onFailedRetryInterval=1,
#     batch=1,
#     duration=0,
#     trimStream=False)
# log(f'Register S OK')

job = {
    'name': 'MegaStar-1to5-S',
    'version': '99.99.99',
    'input_stream_name': create_key_name(Type.STREAM.value, '1MINUTE', Dimension.SECTION.value, '', '', Metric.PAGEVIEWS.value),
    'dimension': Dimension.SECTION.value,
    'number_of_rotated_keys': 5,  # rotate 5 keys of 1 minute each
    'write_to_ts': True,
    'timeseries_name': '5MINUTES',
    'key_expire_duration_sc': 1820,  # keep 6 keys -> merged key expire 30 minutes later
    'batch_size': 10000,
    'batch_interval_ms': 300000,  # run the job every 5 minutes
    'output_stream_name': create_key_name(Type.STREAM.value, '5MINUTES', Dimension.SECTION.value, '', '', Metric.PAGEVIEWS.value)
}

# unregister_old_versions(job.get('name'), job.get('version'))

tracker_log(f'Register {job.get("name")} ...')

desc_json = {
    "name": job.get('name'),
    "version": job.get('version'),
    "desc": f"{job.get('name')} - Rotate Pageviews Keys"
}
GB("StreamReader", desc=json.dumps(desc_json)). \
    aggregate([],
              lambda a, r: a + [r['value']],
              lambda a, r: a + r). \
    map(lambda records: transform(records, job.get('number_of_rotated_keys'))). \
    foreach(lambda records: tracker_log(f'{Dimension.SECTION.value}-map - ids:{len(records.get("id"))} merged:{len(records.get("merge"))} reinjected:{len(records.get("reinject"))}\n{records}')). \
    foreach(lambda records: load(
        job.get('name'),
        records,
        job.get('dimension'),
        job.get('write_to_ts'),
        job.get('timeseries_name'),
        job.get('key_expire_duration_sc'),
        job.get('input_stream_name'),
        job.get('output_stream_name')
    )). \
    register(
    prefix=job.get('input_stream_name'),
    convertToStr=True,
    collect=True,
    onFailedPolicy='abort',
    onFailedRetryInterval=1,
    batch=job.get('batch_size'),
    duration=job.get('batch_interval_ms'),
    trimStream=False)

tracker_log(f'Register {job.get("name")} OK')



from rgtracker.record import *
from rgtracker.tracker import *
from rgtracker.common import *
from rgtracker.website import *
from rgtracker.section import *
from rgtracker.page import *
from rgtracker.device import *
from rgtracker.pageviews import *
from redisgears import log


# Pageviews Rotation Jobs - CMS
pageviews_rotate_jobs = [
    # Run the job every 5 minutes to rotate 5 key of 1 minute each.
    # Expire new merged key after 30 minutes (keep 6 merged keys of 5 minutes each)
    {
        'name': 'MegaStar-1to5-W',
        'version': '99.99.99',
        'input_stream_name': create_key_name(Type.STREAM.value, '1MINUTE', Dimension.WEBSITE.value, '', '', Metric.PAGEVIEWS.value),
        'dimension': Dimension.WEBSITE.value,
        'number_of_rotated_keys': 5, # rotate 5 keys of 1 minute each
        'write_to_ts': True,
        'timeseries_name': '5MINUTES',
        'key_expire_duration_sc': 1820, # keep 6 keys -> merged key expire 30 minutes later
        'batch_size': 10000,
        'batch_interval_ms': 300000, # run the job every 5 minutes
        'output_stream_name': create_key_name(Type.STREAM.value, '5MINUTES', Dimension.WEBSITE.value, '', '', Metric.PAGEVIEWS.value)
    },
    {
        'name': 'MegaStar-1to5-S',
        'version': '99.99.99',
        'input_stream_name': create_key_name(Type.STREAM.value, '1MINUTE', Dimension.SECTION.value, '', '', Metric.PAGEVIEWS.value),
        'dimension': Dimension.SECTION.value,
        'number_of_rotated_keys': 5, # rotate 5 keys of 1 minute each
        'write_to_ts': True,
        'timeseries_name': '5MINUTES',
        'key_expire_duration_sc': 1820, # keep 6 keys -> merged key expire 30 minutes later
        'batch_size': 10000,
        'batch_interval_ms': 300000, # run the job every 5 minutes
        'output_stream_name': create_key_name(Type.STREAM.value, '5MINUTES', Dimension.SECTION.value, '', '', Metric.PAGEVIEWS.value)
    },
    {
        'name': 'MegaStar-1to5-P',
        'version': '99.99.99',
        'input_stream_name': create_key_name(Type.STREAM.value, '1MINUTE', Dimension.PAGE.value, '', '', Metric.PAGEVIEWS.value),
        'dimension': Dimension.PAGE.value,
        'number_of_rotated_keys': 5, # rotate 5 keys of 1 minute each
        'write_to_ts': True,
        'timeseries_name': '5MINUTES',
        'key_expire_duration_sc': 1820, # keep 6 keys -> merged key expire 30 minutes later
        'batch_size': 10000,
        'batch_interval_ms': 300000, # run the job every 5 minutes
        'output_stream_name': create_key_name(Type.STREAM.value, '5MINUTES', Dimension.PAGE.value, '', '', Metric.PAGEVIEWS.value)
    },


    # # Run the job every 10 minutes to rotate 2 key of 5 minutes each.
    # # Expire new merged key after 60 minutes (keep 6 merged keys of 10 minutes each)
    # {
    #     'name': 'MegaStar-5to10-W',
    #     'version': '99.99.99',
    #     'input_stream_name': create_key_name(Type.STREAM.value, '5MINUTES', '', '', '', Metric.PAGEVIEWS.value),
    #     'dimension': Dimension.WEBSITE.value,
    #     'number_of_rotated_keys': 2, # rotate 2 keys of 5 minutes each
    #     'write_to_ts': False,
    #     'timeseries_name': '',
    #     'key_expire_duration_sc': 3620, # keep 6 keys -> merged key expire 60 minutes later
    #     'batch_size': 10000,
    #     'batch_interval_ms': 600000, # run the job every 10 minutes
    #     'output_stream_name': create_key_name(Type.STREAM.value, '10MINUTES', '', '', '', Metric.PAGEVIEWS.value)
    # },
    # {
    #     'name': 'MegaStar-5to10-S',
    #     'version': '99.99.99',
    #     'input_stream_name': create_key_name(Type.STREAM.value, '5MINUTES', '', '', '', Metric.PAGEVIEWS.value),
    #     'dimension': Dimension.SECTION.value,
    #     'number_of_rotated_keys': 2, # rotate 2 keys of 5 minutes each
    #     'write_to_ts': False,
    #     'timeseries_name': '',
    #     'key_expire_duration_sc': 3620, # keep 6 keys -> merged key expire 60 minutes later
    #     'batch_size': 10000,
    #     'batch_interval_ms': 600000, # run the job every 10 minutes
    #     'output_stream_name': create_key_name(Type.STREAM.value, '10MINUTES', '', '', '', Metric.PAGEVIEWS.value)
    # },
    # {
    #     'name': 'MegaStar-5to10-P',
    #     'version': '99.99.99',
    #     'input_stream_name': create_key_name(Type.STREAM.value, '5MINUTES', '', '', '', Metric.PAGEVIEWS.value),
    #     'dimension': Dimension.PAGE.value,
    #     'number_of_rotated_keys': 2, # rotate 2 keys of 5 minutes each
    #     'write_to_ts': False,
    #     'timeseries_name': '',
    #     'key_expire_duration_sc': 3620, # keep 6 keys -> merged key expire 60 minutes later
    #     'batch_size': 10000,
    #     'batch_interval_ms': 600000, # run the job every 10 minutes
    #     'output_stream_name': create_key_name(Type.STREAM.value, '10MINUTES', '', '', '', Metric.PAGEVIEWS.value)
    # },
    # # Run the job every 60 minutes to rotate 6 key of 10 minutes each.
    # # Expire new merged key after 24 hours (keep 24 merged keys of 1 hour each)
    # {
    #     'name': 'MegaStar-10to60-W',
    #     'version': '99.99.99',
    #     'input_stream_name': create_key_name(Type.STREAM.value, '10MINUTES', '', '', '', Metric.PAGEVIEWS.value),
    #     'dimension': Dimension.WEBSITE.value,
    #     'number_of_rotated_keys': 6, # rotate 6 keys of 10 minutes each
    #     'write_to_ts': False,
    #     'timeseries_name': '',
    #     'key_expire_duration_sc': 86420, # keep 6 keys -> merged key expire 60 minutes later
    #     'batch_size': 10000,
    #     'batch_interval_ms': 3600000, # run the job every 60 minutes
    #     'output_stream_name': create_key_name(Type.STREAM.value, '1HOUR', '', '', '', Metric.PAGEVIEWS.value)
    # },
    # {
    #     'name': 'MegaStar-10to60-S',
    #     'version': '99.99.99',
    #     'input_stream_name': create_key_name(Type.STREAM.value, '10MINUTES', '', '', '', Metric.PAGEVIEWS.value),
    #     'dimension': Dimension.SECTION.value,
    #     'number_of_rotated_keys': 6, # rotate 6 keys of 10 minutes each
    #     'write_to_ts': False,
    #     'timeseries_name': '',
    #     'key_expire_duration_sc': 86420, # keep 6 keys -> merged key expire 60 minutes later
    #     'batch_size': 10000,
    #     'batch_interval_ms': 3600000, # run the job every 60 minutes
    #     'output_stream_name': create_key_name(Type.STREAM.value, '1HOUR', '', '', '', Metric.PAGEVIEWS.value)
    # },
    # {
    #     'name': 'MegaStar-10to60-P',
    #     'version': '99.99.99',
    #     'input_stream_name': create_key_name(Type.STREAM.value, '10MINUTES', '', '', '', Metric.PAGEVIEWS.value),
    #     'dimension': Dimension.PAGE.value,
    #     'number_of_rotated_keys': 6, # rotate 6 keys of 10 minutes each
    #     'write_to_ts': False,
    #     'timeseries_name': '',
    #     'key_expire_duration_sc': 86420, # keep 6 keys -> merged key expire 60 minutes later
    #     'batch_size': 10000,
    #     'batch_interval_ms': 3600000, # run the job every 60 minutes
    #     'output_stream_name': create_key_name(Type.STREAM.value, '1HOUR', '', '', '', Metric.PAGEVIEWS.value)
    # },
]
for job in pageviews_rotate_jobs:
    # unregister_old_versions(job.get('name'), job.get('version'))
    desc_json = {
        "name": job.get('name'),
        "version": job.get('version'),
        "desc": f"{job.get('name')} - Rotate Pageviews Keys"
    }
    tracker_log(f'Register {job.get("name")} ...')
    GB("StreamReader", desc=json.dumps(desc_json)). \
        foreach(lambda records: tracker_log(f'start - {len(records)}->{records}')). \
        aggregate([],
                  lambda a, r: a + [r['value']],
                  lambda a, r: a + r). \
        map(lambda records: transform(records, job.get('number_of_rotated_keys'))). \
        foreach(lambda records: tracker_log(f'map - {records}')). \
        foreach(lambda records: load(
            records,
            job.get('dimension'),
            job.get('write_to_ts'),
            job.get('timeseries_name'),
            job.get('key_expire_duration_sc'),
            job.get('input_stream_name'),
            job.get('output_stream_name')
        )). \
        register(
        prefix=job.get('input_stream_name'),
        convertToStr=True,
        collect=True,
        onFailedPolicy='abort',
        onFailedRetryInterval=1,
        batch=job.get('batch_size'),
        duration=job.get('batch_interval_ms'),
        trimStream=False)
    tracker_log(f'Register {job.get("name")} OK')



from rgtracker.record import *
from rgtracker.tracker import *
from rgtracker.common import *
from rgtracker.website import *
from rgtracker.section import *
from rgtracker.page import *
from rgtracker.device import *
from rgtracker.pageviews import *
import inspect

job = {
    'name': 'W-PG-1to5',
    'version': '99.99.99',
    'input_stream_name': create_key_name(Type.STREAM.value, '1MINUTE', Dimension.WEBSITE.value, '', '', Metric.PAGEVIEWS.value),
    'dimension': Dimension.WEBSITE.value,
    'number_of_rotated_keys': 5, # rotate 5 keys of 1 minute each
    'write_to_ts': True,
    'timeseries_name': '5MINUTES',
    # 'key_expire_duration_sc': 1820, # keep 6 keys -> merged key expire 30 minutes later
    'key_expire_duration_sc': 3600, # TEST
    'batch_size': 9999999,
    'batch_interval_ms': 300000, # run the job every 5 minutes
    'output_stream_name': create_key_name(Type.STREAM.value, '5MINUTES', Dimension.WEBSITE.value, '', '', Metric.PAGEVIEWS.value)
}

# Todo: unregister_old_versions job
# unregister_old_versions(job.get('name'), job.get('version'))

tracker_log(f'Register {job.get("name")} ...')

desc_json = {
    "name": job.get('name'),
    "version": job.get('version'),
    "desc": f"Rotate Website Pageviews Keys, 1 minute to 5 minutes buckets"
}


def load_fix(job_name, records, dimension, write_to_ts, timeseries_name, key_expire_duration_sc, reinject_stream_name,
             output_stream_name):
    return load(
        job_name,
        records,
        dimension,
        write_to_ts,
        timeseries_name,
        key_expire_duration_sc,
        reinject_stream_name,
        output_stream_name
    )


GB("StreamReader", desc=json.dumps(desc_json)). \
    aggregate([],
              lambda a, r: a + [r['value']],
              lambda a, r: a + r). \
    map(lambda records: Pageviews.transform(records=records, number_of_rotated_keys=job.get('number_of_rotated_keys'), dimension=job.get('dimension'))). \
    foreach(lambda x: tracker_log(f'{inspect.signature(Pageviews.load)}', f'DEBUG - {job.get("name")} - ')). \
    foreach(lambda records: Pageviews.load(
        job_name=job.get('name'),
        records=records,
        dimension=job.get('dimension'),
        write_to_ts=job.get('write_to_ts'),
        timeseries_name=job.get('timeseries_name'),
        key_expire_duration_sc=job.get('key_expire_duration_sc'),
        reinject_stream_name=job.get('input_stream_name'),
        output_stream_name=job.get('output_stream_name')
    )). \
    foreach(lambda records: tracker_log(f'ids:{len(records.get("ids"))} merged:{len(records.get("merge"))} reinjected:{len(records.get("reinject"))} - {records}', f'{job.get("name")} - ')). \
    register(
        prefix=job.get('input_stream_name'),
        convertToStr=True,
        collect=True,
        onFailedPolicy='abort',
        onFailedRetryInterval=1,
        batch=job.get('batch_size'),
        duration=job.get('batch_interval_ms'),
        trimStream=False)

tracker_log(f'Register {job.get("name")} OK')



from rgtracker.record import *
from rgtracker.tracker import *
from rgtracker.common import *
from rgtracker.website import *
from rgtracker.section import *
from rgtracker.page import *
from rgtracker.device import *
from rgtracker.pageviews import *
from rgtracker.uniquedevices import *
from redisgears import executeCommand as execute

def compare_versions_sandbox(v1, v2):
    # None version is less then all version
    if v1 is None:
        return -1
    if v2 is None:
        return 1

    if v1 == "99.99.99":
        return 1
    if v2 == "99.99.99":
        return -1

    v1_major, v1_minor, v1_patch = v1.split(".")
    v2_major, v2_minor, v2_patch = v2.split(".")

    if int(v1_major) > int(v2_major):
        return 1
    elif int(v1_major) < int(v2_major):
        return -1

    if int(v1_minor) > int(v2_minor):
        return 1
    elif int(v1_minor) < int(v2_minor):
        return -1

    if int(v1_patch) > int(v2_patch):
        return 1
    elif int(v1_patch) < int(v2_patch):
        return -1

    return 0

def unregister_old_versions_sandbox(name, version):
    tracker_log(f"Unregister old version of {name}")
    registrations = execute("RG.DUMPREGISTRATIONS")
    for registration in registrations:
        registration_dict = registration_arr_to_dict(registration, 0)
        desc_str = registration_dict["desc"]
        try:
            desc = json.loads(desc_str)
        except Exception as e:
            continue
        if "name" in desc.keys() and name in desc["name"]:
            tracker_log(
                "Version auto upgrade is not atomic, make sure to use it when is not traffic to the database (otherwise you might lose events).",
                log_level="warning"
            )
            if "version" not in desc.keys():
                execute("RG.UNREGISTER", registration_dict["id"])
                tracker_log(f"Unregister {registration_dict['id']}")
                continue
            v = desc["version"]
            if compare_versions_sandbox(version, v) > 0:
                execute("RG.UNREGISTER", registration_dict["id"])
                tracker_log(f"Unregistered {registration_dict['id']}")
            else:
                raise Exception(
                    "Found a version which is greater or equals current version, aborting."
                )
    tracker_log("Unregistered old versions.")

# unregister_old_versions_sandbox("StreamReaderJob", '99.99.99')

descJson = {
    "name": f"StreamReaderJob",
    "version": '0.0.2',
    "desc": f"Lorem Ipsum",
}

unregister_old_versions_sandbox(descJson.get('name'), descJson.get('version'))

GB("StreamReader", desc=json.dumps(descJson)). \
    foreach(lambda record: tracker_log(f'{record}', f'{descJson.get("name")} - {descJson.get("version")} - ')). \
    register(
    prefix=f'ST:TRACKER::::',
    convertToStr=True,
    collect=True,
    onFailedPolicy='abort',
    onFailedRetryInterval=1,
    batch=1,
    duration=0,
    trimStream=False)



from rgtracker.tracker import *
from rgtracker.website import *
from rgtracker.section import *
from rgtracker.page import *
from dataclasses import dataclass, field


@dataclass
class Record:
    """
    Represent any records read with Redis Gears

    Examples
    --------
    Record.key = TYPE:optional(NAME):optional(DIMENSION):optional(RECORD_ID):optional(DATETIME):optional(METRIC)
    """
    redis_event: str
    redis_key: str
    redis_type: str
    redis_value: dict
    type: field(init=False) = None
    name: field(init=False) = None
    dimension: field(init=False) = None
    record_id: field(init=False) = None
    ts: field(init=False) = None
    metric_id: field(init=False) = None

    def __post_init__(self):
        key_split = self.redis_key.split(':')
        self.type = self.check_none(key_split[RedisNC.TYPE])
        self.name = self.check_none(key_split[RedisNC.NAME])
        self.dimension = self.check_none(key_split[RedisNC.DIMENSION])
        self.record_id = self.check_none(key_split[RedisNC.RECORD_ID])
        self.ts = self.check_none(key_split[RedisNC.TS])
        self.metric = self.check_none(key_split[RedisNC.METRIC])

    @staticmethod
    def check_none(check_value):
        if check_value == '':
            return None
        else:
            return check_value

    def serialize_record(self):
        # Todo: use from_dict
        if self.type == Type.STREAM.value:
            # Todo: check schema version with metadata field and create 'Stream' object
            if self.name == 'TRACKER':
                return Tracker(
                    uui=self.redis_value.get('uui'),
                    url=self.redis_value.get('url'),
                    ts=self.redis_value.get('dt'),
                    metadata=self.redis_value.get('metadata'),
                )
            elif self.name.startswith('BLOOM-ROTATE'):
                if self.redis_value.get('dimension') == Dimension.WEBSITE.value:
                    x = Website(
                        id=self.redis_value.get('key').split(':')[RedisNC.RECORD_ID],
                        last_visited=self.redis_value.get('key').split(':')[RedisNC.TS]
                    )
                    return x
                elif self.redis_value.get('dimension') == Dimension.SECTION.value:
                    return Section(
                        id=self.redis_value.get('key').split(':')[RedisNC.RECORD_ID],
                        last_visited=self.redis_value.get('key').split(':')[RedisNC.TS]
                    )
                elif self.redis_value.get('dimension') == Dimension.PAGE.value:
                    return Page(
                        id=self.redis_value.get('key').split(':')[RedisNC.RECORD_ID],
                        last_visited=self.redis_value.get('key').split(':')[RedisNC.TS]
                    )
                else:
                    raise ValueError
            else:
                raise ValueError
        else:
            if self.dimension == Dimension.WEBSITE.value:
                raise NotImplementedError
            elif self.dimension == Dimension.SECTION.value:
                # Todo: Use from_dict method
                # return from_dict(data_class=Section, data={
                #     'id': self.record_id,
                #     'name': None  # self.get_dimension_value()
                # })
                raise NotImplementedError
            elif self.dimension == Dimension.PAGE.value:
                raise NotImplementedError
            elif self.dimension == Dimension.DEVICE.value:
                raise NotImplementedError
            else:
                raise ValueError



from rgtracker.common import *
import math
import pandas as pd
from redisgears import executeCommand as execute
from redisgears import log


def extract(record, dimension):
    return record['value']['dimension'] == dimension


def transform(records, number_of_rotated_keys, dimension):
    df = pd.DataFrame(records)
    df['ids'] = df['ids'].apply(lambda x: json.loads(x) if x.startswith('[') else x)
    # tracker_log(f'all_ids - {df["ids"].values.tolist()}', f'W-UD-1to5 - transform - ')

    def flatten_list(A):
        rt = []
        for i in A:
            if isinstance(i, list):
                rt.extend(flatten_list(i))
            else:
                rt.append(i)
        return rt

    def unique_item(_list):
        return list(set(_list))

    grouped_df = df.groupby('ts').agg(lambda ids: unique_item(flatten_list(list(ids)))).sort_values('ts').reset_index()
    # tracker_log(f'all_ts - {grouped_df["ts"].values.tolist()}', f'W-PG-1to5 - transform - ')

    expected_rows = number_of_rotated_keys
    chunks = math.floor(len(grouped_df['ts']) / expected_rows + 1)
    i = 0
    j = expected_rows

    results = {
        'ids': [],
        'merge': [],
        'reinject': []
    }
    for x in range(chunks):
        df_sliced = grouped_df[i:j]
        if df_sliced.shape[0] >= number_of_rotated_keys:

            # Add only ids from 'complete' chunk
            ts = df_sliced["ts"].values.tolist()
            # tracker_log(f'chunk_{x}_ts - {ts}', f'X-PG-1to5 - transform - ')

            ids = df_sliced.apply(lambda ids: unique_item(flatten_list(ids)))['ids']
            for unique_id in ids:
                results.get('ids').append(unique_id)
            # tracker_log(f'chunk_{x}_ids - {ids}', f'X-PG-1to5 - transform - ')

            results.get('merge').append({
                'name': create_key_name(
                    Type.HLL.value,
                    '',
                    dimension,
                    '',
                    get_ts_df(ts[0], df_sliced["ts"]),
                    Metric.UNIQUE_DEVICES.value
                ),
                'keys': [create_key_name(Type.HLL.value, '', dimension, '', ts, Metric.UNIQUE_DEVICES.value) for ts in
                         df_sliced['ts']]
            })
        else:
            [results.get('reinject').append({
                'ts': row[1],
                'ids': row[0]
            }) for row in zip(df_sliced['ids'], df_sliced['ts'])]

        i += expected_rows
        j += expected_rows

    # Deduplicate ids
    results.update({'ids': list(set(results.get('ids')))})
    # tracker_log(f'final_ids - {results.get("ids")}', f'X-PG-1to5 - transform - ')
    return results


def load(job_name, records, dimension, write_to_ts, timeseries_name, key_expire_duration_sc, reinject_stream_name,
         output_stream_name):
    def get_ts(ts):
        if len(ts.split('_')) > 1:
            return ts.split("_")[-1]
        else:
            return ts

    # tracker_log(
    #     f'ids:{len(records.get("ids"))} merged:{len(records.get("merge"))} reinjected:{len(records.get("reinject"))}\n{records}',
    #     f'{job_name} - load - ')

    capped_stream = get_maxlen_capped_stream(timeseries_name, dimension)

    index_name = create_key_name(
        type=Type.INDEX.value,
        name='',
        dimension=dimension,
        record_id='',
        ts='',
        metric='')

    for hll_reinject in records.get('reinject'):
        execute('XADD', reinject_stream_name, 'MAXLEN', f'{capped_stream}', '*',
                'ts', hll_reinject.get('ts'),
                'status', 'reinject',
                'ids', json.dumps(hll_reinject.get('ids')))
        # tracker_log(
        #     f'XADD {reinject_stream_name} * ts {hll_reinject.get("ts")} status reinject ids {json.dumps(hll_reinject.get("ids"))}',
        #     f'{job_name} - load - ')

    for hll_merge in records.get("merge"):
        # FixMe: useless EXISTS, merged key never exists. Example of merged key: HLL::W::1663581720000_1663582260000:UD
        # Move this into 'ids' loop?
        if execute('EXISTS', hll_merge.get('name')) != 1:

            parsed_key = parse_key_name(hll_merge.get('name'))
            for id in records.get('ids'):
                k = create_key_name(Type.HLL.value, '', dimension, id, parsed_key.get('ts'),
                                    Metric.UNIQUE_DEVICES.value)
                keys = [create_key_name(Type.HLL.value, '', dimension, id, parse_key_name(k).get('ts'),
                                        Metric.UNIQUE_DEVICES.value) for k in hll_merge.get('keys')]
                execute('PFMERGE', k, *keys)
                # tracker_log(f'PFMERGE {k} {keys}', f'{job_name} - load - ')

                if write_to_ts:
                    unique_devices = execute('PFCOUNT', *keys)
                    # tracker_log(f'PFCOUNT {keys}', f'{job_name} - load - ')

                    timeseries_key_name = create_key_name(
                        type=Type.TIMESERIES.value,
                        name=timeseries_name,
                        dimension=dimension,
                        record_id=id,
                        metric=Metric.UNIQUE_DEVICES.value)

                    if dimension == Dimension.WEBSITE.value:
                        # record_infos = execute('FT.SEARCH', index_name, f'@id:{{{id}}}',
                        #                        'RETURN', '3', 'name', 'AS', 'website_name')
                        # tracker_log(f'Get {dimension} infos - {record_infos}', f'{job_name} ')
                        json_key = create_key_name(Type.JSON.value, '', Dimension.WEBSITE.value, id, '', '')
                        # JSON.GET J::S:46535707faf945b66a11ddf37f8017b2c29bfa3e:: website.id website.name name
                        record_infos = execute('JSON.GET', json_key, 'name')
                        record_infos = json.loads(record_infos)

                        # FixMe: HOT FIX -> no records info found
                        if bool(record_infos):
                            execute('TS.ADD', timeseries_key_name, get_ts(parsed_key.get('ts')), unique_devices,
                                    'ON_DUPLICATE', 'LAST',
                                    'LABELS', 'ts_name', timeseries_name,
                                    'dimension', dimension, Dimension.METRIC.value, Metric.UNIQUE_DEVICES.value,
                                    'website_id', id, 'website_name', record_infos)
                            tracker_log(
                                f'Write {unique_devices} devices to {timeseries_key_name} at {get_ts(parsed_key.get("ts"))}',
                                f'{job_name} - ')
                        else:
                            tracker_log(
                                f'Write {unique_devices} devices to {timeseries_key_name} at {get_ts(parsed_key.get("ts"))}',
                                f'{job_name} - warning - ')
                    elif dimension == Dimension.SECTION.value:
                        # record_infos = execute('FT.SEARCH', index_name, f'@id:{{{id}}}', 'RETURN', '5',
                        #                        'name', 'AS', 'section_name', 'website_id', 'website_name')
                        json_key = create_key_name(Type.JSON.value, '', Dimension.SECTION.value, id, '', '')
                        # JSON.GET J::S:46535707faf945b66a11ddf37f8017b2c29bfa3e:: website.id website.name name
                        record_infos = execute('JSON.GET', json_key, 'website.id', 'website.name', 'name')
                        record_infos = json.loads(record_infos)

                        # tracker_log(f'Get {dimension} infos - {record_infos}', f'{job_name} ')
                        # FixMe: HOT FIX -> no records info found
                        if bool(record_infos):
                            execute('TS.ADD', timeseries_key_name, get_ts(parsed_key.get("ts")), unique_devices,
                                    'ON_DUPLICATE', 'LAST',
                                    'LABELS', 'ts_name', timeseries_name,
                                    'dimension', dimension, Dimension.METRIC.value, Metric.UNIQUE_DEVICES.value,
                                    'section_id', id, 'section_name', record_infos.get('name'),
                                    'website_id', record_infos.get('website.id'), 'website_name', record_infos.get('website.name'))
                            tracker_log(
                                f'Write {unique_devices} devices to {timeseries_key_name} at {get_ts(parsed_key.get("ts"))}',
                                f'{job_name} - ')
                        else:
                            tracker_log(
                                f'Write {unique_devices} devices to {timeseries_key_name} at {get_ts(parsed_key.get("ts"))}',
                                f'{job_name} - warning - ')
                    elif dimension == Dimension.PAGE.value:
                        # record_infos = execute('FT.SEARCH', index_name, f'@id:{{{id}}}', 'RETURN', '5',
                        #                        'website_id', 'website_name', 'section_id', 'section_name', 'article_id')
                        json_key = create_key_name(Type.JSON.value, '', Dimension.PAGE.value, id, '', '')
                        # JSON.GET J::P:620070307ea50a61c177f8c3dbb063b708dad92c:: website.id website.name section.id section.name article_id
                        record_infos = execute('JSON.GET', json_key, 'website.id', 'website.name', 'section.id', 'section.name', 'article_id')
                        record_infos = json.loads(record_infos)

                        # tracker_log(f'Get {dimension} infos - {record_infos}', f'{job_name} ')
                        if bool(record_infos):
                            execute('TS.ADD', timeseries_key_name, get_ts(parsed_key.get("ts")), unique_devices,
                                    'ON_DUPLICATE', 'LAST',
                                    'LABELS', 'ts_name', timeseries_name,
                                    'dimension', dimension, Dimension.METRIC.value, Metric.UNIQUE_DEVICES.value,
                                    'page_id', id,
                                    'website_id', record_infos.get('website.id'), 'website_name', record_infos.get('website.name'),
                                    'section_id', record_infos.get('section.id'), 'section_name', record_infos.get('section.name'),
                                    'article_id', record_infos.get('article_id'))
                            tracker_log(
                                f'Write {unique_devices} devices to {timeseries_key_name} at {get_ts(parsed_key.get("ts"))}',
                                f'{job_name} - ')
                        else:
                            tracker_log(
                                f'Write {unique_devices} devices to {timeseries_key_name} at {get_ts(parsed_key.get("ts"))}',
                                f'{job_name} - warning - ')

                execute('EXPIRE', k, key_expire_duration_sc)

            execute('XADD', output_stream_name, 'MAXLEN', f'{capped_stream}', '*',
                    'ts', parsed_key.get('ts'),
                    'status', 'merged',
                    'ids', json.dumps(records.get('ids')))
            # tracker_log(f'Write key {hll_merge.get("name")} with ids {json.dumps(records.get("ids"))} to {output_stream_name} ', f'{job_name} - ')



    return records



from rgtracker.website import *
from rgtracker.section import *
from rgtracker.page import *
from rgtracker.device import *
import hashlib
from dataclasses import dataclass, field
from pathlib import PurePosixPath
from urllib.parse import unquote, urlparse


@dataclass
class Tracker:
    """Represent Main object in Tracker project. Come from ST:TRACKER:::: Stream"""
    uui: str
    url: str
    ts: str
    metadata: dict
    website: field(init=False) = None
    section: field(init=False) = None
    page: field(init=False) = None

    def __post_init__(self):
        x = self.harmonize_url(self.url)
        self.website = x.get('website')
        self.section = x.get('section')
        self.page = x.get('page')

    @staticmethod
    def generate_id(value):
        return hashlib.sha1(str(value).encode('utf-8')).hexdigest()

    def harmonize_url(self, url):
        o = urlparse(url)

        website = o.hostname
        if website == 'www.rtl.lu':
            website = 'rtl'
        elif website == '5minutes.rtl.lu':
            website = '5minutes'
        elif website == 'today.rtl.lu':
            website = 'today'
        elif website == 'www.eldo.lu':
            website = 'eldo'
        else:
            website = "other"

        level_values = list(PurePosixPath(unquote(o.path)).parts)
        level_values.append(level_values.pop(-1).split('.')[0])

        level_ids = ["level_{}".format(i) for i in range(len(level_values))]
        levels = dict(zip(level_ids, level_values))

        section = None
        if len(level_values) > 3:
            section = f"{level_values[1]}/{level_values[2]}"
        else:
            section = f"{level_values[1]}"

        clean_url = o.netloc + o.path
        article_id = int(level_values[-1]) if level_values[-1].isdigit() else None

        return {
            'website': {
                'id': self.generate_id(website),
                'name': website,
                'last_visited': self.ts
            },
            'section': {
                'id': self.generate_id(section),
                'name': section,
                'levels': levels,
                'last_visited': self.ts
            },
            'page': {
                'id': self.generate_id(clean_url),
                'url': clean_url,
                'last_visited': self.ts,
                'article_id': article_id,
            }
        }

    def serialize_tracker(self):
        return {
            'website': Website(
                id=self.website.get('id'),
                name=self.website.get('name'),
                last_visited=self.website.get('last_visited')
            ),
            'section': Section(
                id=self.section.get('id'),
                name=self.section.get('name'),
                levels=dict(self.section.get('levels')),
                last_visited=self.section.get('last_visited')
            ),
            'page': Page(
                id=self.page.get('id'),
                url=self.page.get('url'),
                article_id=self.page.get('article_id'),
                last_visited=self.page.get('last_visited')
            ),
            'device': Device(id=self.uui)
        }



from enum import Enum, IntEnum, unique
from redis import Redis
from redis.commands.search.field import (
    NumericField,
    TagField,
    TextField,
)
from redis.commands.search.indexDefinition import IndexDefinition, IndexType


@unique
class RedisNC(IntEnum):
    TYPE = 0,
    NAME = 1,
    DIMENSION = 2,
    RECORD_ID = 3,
    TS = 4,
    METRIC = 5


@unique
class Type(Enum):
    STREAM = 'ST'
    HASH = 'H'
    JSON = 'J'
    INDEX = 'I'
    TIMESERIES = 'TS'
    BLOOM = 'B'
    SORTEDSET = 'SS'
    SET = 'S'
    LIST = 'L'
    CHANNEL = 'C'
    CMS = 'CMS'
    HLL = 'HLL'


@unique
class Dimension(Enum):
    WEBSITE = 'W'
    SECTION = 'S'
    PAGE = 'P'
    DEVICE = 'D'
    AUDIO = 'A'
    VIDEO = 'V'
    PODCAST = 'PC'
    METRIC = 'M'


@unique
class Metric(Enum):
    PAGEVIEWS = 'PG'
    DEVICES = 'D'
    UNIQUE_DEVICES = 'UD'


def create_key_name(type, name='', dimension='', record_id='', ts='', metric=''):
    return f'{type}:{name}:{dimension}:{record_id}:{ts}:{metric}'


def create_website_index(redis):
    try:
        index_name = create_key_name(
            type=Type.INDEX.value, name='',
            dimension=Dimension.WEBSITE.value, record_id='',
            ts='', metric='')
        schema = (
            TagField('$.id', as_name='id'),
            TextField('$.name', as_name='name'),
            NumericField("$.last_visited", as_name='last_visited', sortable=True),
            TagField(name='$.sections[*].id', as_name='section_id'),
            TagField('$.sections[*].name', as_name='section_name', separator='/'),
            # NumericField('$.sections[*].last_visited', as_name='section_last_visited', sortable=True),
            TagField(name='$.pages[*].id', as_name='page_id'),
            # TagField('$.pages[0:].url', as_name='page_url'),
            # TagField('$.pages[*].article_id', as_name='page_article_id'),
            # NumericField('$.pages[0:].last_visited', as_name='page_last_visited', sortable=True),
        )
        definition = IndexDefinition(prefix=['J::W:'], index_type=IndexType.JSON)
        redis.ft(index_name).create_index(schema, definition=definition)
    except Exception as e:
        print(f'Error when creating Redis index: {e}')
        pass


def create_section_index(redis):
    try:
        index_name = create_key_name(
            type=Type.INDEX.value,
            name='',
            dimension=Dimension.SECTION.value,
            record_id='',
            ts='',
            metric='')
        schema = (
            TagField('$.id', as_name='id'),
            TagField('$.name', as_name='name', separator='/'),
            TagField('$.level_0', as_name='level_0'),
            TagField('$.level_1', as_name='level_1'),
            TagField('$.level_2', as_name='level_2'),
            TagField('$.level_3', as_name='level_3'),
            TagField('$.level_4', as_name='level_4'),
            NumericField("$.last_visited", as_name='last_visited', sortable=True),
            TagField(name='$.website.id', as_name='website_id'),
            TagField('$.website.name', as_name='website_name'),
        )
        definition = IndexDefinition(prefix=['J::S:'], index_type=IndexType.JSON)
        redis.ft(index_name).create_index(schema, definition=definition)
    except Exception as e:
        print(f'Error when creating Redis index: {e}')
        pass


def create_page_index(redis):
    try:
        index_name = create_key_name(
            type=Type.INDEX.value, name='',
            dimension=Dimension.PAGE.value, record_id='',
            ts='', metric='')
        schema = (
            TagField('$.id', as_name='id'),
            TextField('$.url', as_name='url'),
            # TagField('$.article_id', as_name='article_id'), # Tag field doesn't accept int field -> str(article_id)
            TextField('$.metadata.title', as_name='title'),
            TextField('$.metadata.kicker', as_name='kicker'),
            NumericField("$.last_visited", as_name='last_visited', sortable=True),
            TagField(name='$.website.id', as_name='website_id'),
            TagField('$.website.name', as_name='website_name'),
            TagField(name='$.section.id', as_name='section_id'),
            TagField('$.section.name', as_name='section_name', separator='/'),
            TagField('$.section.levels.level_0', as_name='section_level_0'),
            TagField('$.section.levels.level_1', as_name='section_level_1'),
            TagField('$.section.levels.level_2', as_name='section_level_2'),
            TagField('$.section.levels.level_3', as_name='section_level_3'),
            TagField('$.section.levels.level_4', as_name='section_level_4'),
        )
        definition = IndexDefinition(prefix=['J::P:'], index_type=IndexType.JSON)
        redis.ft(index_name).create_index(schema, definition=definition)
    except Exception as e:
        print(f'Error when creating Redis index: {e}')
        pass


redis_conn = Redis(host="localhost", port=6379)
# redis_conn = Redis(host="10.61.228.218", port=6379, password='PiPaPo-2021.12')

create_website_index(redis_conn)
create_section_index(redis_conn)
create_page_index(redis_conn)



from rgtracker.common import *
from rgtracker.record import *
from rgtracker.tracker import *
from rgtracker.website import *
from rgtracker.section import *
from rgtracker.page import *
from rgtracker.device import *


def transform(record):
    serialized_object = Record(
        redis_event=record.get('event'),
        redis_key=record.get('key'),
        redis_type=record.get('type'),
        redis_value=record.get('value')
    ).serialize_record()

    if type(serialized_object) is Tracker:
        return serialized_object.serialize_tracker()
    else:
        raise Exception(f'Record is not a Tracker object')


def load(tracker):

    # Todo: TopK hitter! (Top 50 for each minute)

    website: Website = tracker.get('website')
    section = tracker.get('section')
    page = tracker.get('page')
    device = tracker.get('device')

    output_stream_names = {
        'default': create_key_name(Type.STREAM.value, '1MINUTE'),
        'pageviews': create_key_name(Type.STREAM.value, '1MINUTE', '', '', '', Metric.PAGEVIEWS.value),
        'unique_devices': create_key_name(Type.STREAM.value, '1MINUTE', '', '', '', Metric.UNIQUE_DEVICES.value),
        'website_pageviews': create_key_name(Type.STREAM.value, '1MINUTE', Dimension.WEBSITE.value, '', '', Metric.PAGEVIEWS.value),
        'section_pageviews': create_key_name(Type.STREAM.value, '1MINUTE', Dimension.SECTION.value, '', '', Metric.PAGEVIEWS.value),
        'page_pageviews': create_key_name(Type.STREAM.value, '1MINUTE', Dimension.PAGE.value, '', '', Metric.PAGEVIEWS.value),
        'website_unique_devices': create_key_name(Type.STREAM.value, '1MINUTE', Dimension.WEBSITE.value, '', '', Metric.UNIQUE_DEVICES.value),
        'section_unique_devices': create_key_name(Type.STREAM.value, '1MINUTE', Dimension.SECTION.value, '', '', Metric.UNIQUE_DEVICES.value),
        'page_unique_devices': create_key_name(Type.STREAM.value, '1MINUTE', Dimension.PAGE.value, '', '', Metric.UNIQUE_DEVICES.value),
    }
    if not is_weird(website.last_visited, website.last_visited_rounded):
        load_website(website, section, page, device, output_stream_names)
        load_section(website, section, page, device, output_stream_names)
        load_page(website, section, page, device, output_stream_names)

    return tracker


desc_json = {
    "name": 'BigBang',
    "version": '99.99.99',
    "desc": f"BigBang - Backbone Job"
}
# unregister_old_versions(desc_json.get('name'), desc_json.get('version'))
GB("StreamReader", desc=json.dumps(desc_json)). \
    map(transform). \
    foreach(load). \
    register(
    prefix='ST:TRACKER::::',
    convertToStr=True,
    collect=True,
    onFailedPolicy='abort',
    onFailedRetryInterval=1,
    batch=1,
    duration=0,
    trimStream=False)

# run('ST:TRACKER::::', trimStream=False)
# register(
# prefix='ST:TRACKER::::',
# convertToStr=True,
# collect=True,
# onFailedPolicy='abort',
# onFailedRetryInterval=1,
# batch=1,
# duration=0,
# trimStream=False)



from rgtracker.common import *
import math
import pandas as pd
from redisgears import executeCommand as execute
from dataclasses import dataclass, field


@dataclass
class Pageviews:
    r: field(init=False) = None
    d: field(init=False) = None

    def extract(self, record, dimension):
        tracker_log(f'{self.r} {self.d}', f'Pageviews - ')
        return record['value']['dimension'] == dimension

    def transform(self, records, number_of_rotated_keys, dimension):
        tracker_log(f'{self.r} {self.d}', f'Pageviews - ')
        df = pd.DataFrame(records)
        df['ids'] = df['ids'].apply(lambda x: json.loads(x) if x.startswith('[') else x)

        # tracker_log(f'all_ids - {df["ids"].values.tolist()}', f'W-PG-1to5 - transform - ')

        def flatten_list(A):
            rt = []
            for i in A:
                if isinstance(i, list):
                    rt.extend(flatten_list(i))
                else:
                    rt.append(i)
            return rt

        def unique_item(_list):
            return list(set(_list))

        grouped_df = df.groupby('ts').agg(lambda ids: unique_item(flatten_list(list(ids)))).sort_values(
            'ts').reset_index()
        # tracker_log(f'all_ts - {grouped_df["ts"].values.tolist()}', f'W-PG-1to5 - transform - ')

        expected_rows = number_of_rotated_keys
        chunks = math.floor(len(grouped_df['ts']) / expected_rows + 1)
        i = 0
        j = expected_rows

        results = {
            'ids': [],
            'merge': [],
            'reinject': []
        }
        for x in range(chunks):
            df_sliced = grouped_df[i:j]
            if df_sliced.shape[0] >= number_of_rotated_keys:

                # Add only ids from 'complete' chunk
                ts = df_sliced["ts"].values.tolist()
                # tracker_log(f'chunk_{x}_ts - {ts}', f'X-PG-1to5 - transform - ')

                ids = df_sliced.apply(lambda ids: unique_item(flatten_list(ids)))['ids']
                for unique_id in ids:
                    results.get('ids').append(unique_id)
                # tracker_log(f'chunk_{x}_ids - {ids}', f'X-PG-1to5 - transform - ')

                results.get('merge').append({
                    'name': create_key_name(
                        Type.CMS.value,
                        '',
                        dimension,
                        '',
                        get_ts_df(ts[0], df_sliced["ts"]),
                        Metric.PAGEVIEWS.value
                    ),
                    'keys': [create_key_name(Type.CMS.value, '', dimension, '', ts, Metric.PAGEVIEWS.value) for ts in
                             df_sliced['ts']]
                })
            else:
                [results.get('reinject').append({
                    'ts': row[1],
                    'ids': row[0]
                }) for row in zip(df_sliced['ids'], df_sliced['ts'])]

            i += expected_rows
            j += expected_rows

        # Deduplicate ids
        results.update({'ids': list(set(results.get('ids')))})
        # tracker_log(f'final_ids - {results.get("ids")}', f'X-PG-1to5 - transform - ')
        return results

    def load(self, job_name, records, dimension, write_to_ts, timeseries_name, key_expire_duration_sc,
             reinject_stream_name,
             output_stream_name):
        def get_ts(ts):
            if len(ts.split('_')) > 1:
                return ts.split("_")[-1]
            else:
                return ts

        capped_stream = get_maxlen_capped_stream(timeseries_name, dimension)

        index_name = create_key_name(
            type=Type.INDEX.value,
            name='',
            dimension=dimension,
            record_id='',
            ts='',
            metric='')

        for cms_reinject in records.get('reinject'):
            execute('XADD', reinject_stream_name, 'MAXLEN', f'{capped_stream}', '*',
                    'ts', cms_reinject.get('ts'),
                    'status', 'reinject',
                    'ids', json.dumps(cms_reinject.get('ids')))
            # tracker_log(f'Reinject {cms_reinject}', f'{job_name} - ')

        for cms_merge in records.get("merge"):
            # tracker_log(f'Merge {cms_merge}', f'{job_name} - ')

            if execute('EXISTS', cms_merge.get('name')) != 1:
                try:
                    execute('CMS.INITBYDIM', cms_merge.get('name'), 2000, 5)
                except Exception as e:
                    tracker_log(f'Error during CMS key init: {e} {cms_merge.get("name")}', f'{job_name} - ', 'warning')

                try:
                    execute('CMS.MERGE', cms_merge.get('name'), len(cms_merge.get('keys')), *cms_merge.get('keys'))
                except Exception as e:
                    tracker_log(
                        f'Error during CMS merging key: {e} {cms_merge.get("name")} {len(cms_merge.get("keys"))} {cms_merge.get("keys")}',
                        f'{job_name} ', 'warning')
                    # FixMe: one of merged keys is expire or missing, create strategy to not crash app and standardize data

                parsed_key_name = parse_key_name(cms_merge.get('name'))

                if write_to_ts:
                    for id in records.get('ids'):
                        pageviews = execute('CMS.QUERY', cms_merge.get('name'), id)[0]

                        timeseries_key_name = create_key_name(
                            type=Type.TIMESERIES.value,
                            name=timeseries_name,
                            dimension=dimension,
                            record_id=id,
                            metric=Metric.PAGEVIEWS.value)

                        if dimension == Dimension.WEBSITE.value:
                            # record_infos = execute('FT.SEARCH', index_name, f'@id:{{{id}}}', 'RETURN', '3', 'name', 'AS',
                            #                        'website_name')
                            json_key = create_key_name(Type.JSON.value, '', Dimension.WEBSITE.value, id, '', '')
                            # JSON.GET J::S:46535707faf945b66a11ddf37f8017b2c29bfa3e:: website.id website.name name
                            record_infos = execute('JSON.GET', json_key, 'name')
                            record_infos = json.loads(record_infos)

                            # tracker_log(f'records_infos {json_key} {record_infos}', f'{job_name} - ')
                            # FixMe: HOT FIX -> no records info found
                            if bool(record_infos):
                                execute('TS.ADD', timeseries_key_name, get_ts(parsed_key_name.get('ts')), pageviews,
                                        'ON_DUPLICATE', 'LAST',
                                        'LABELS', 'ts_name', timeseries_name,
                                        'dimension', dimension, Dimension.METRIC.value, Metric.PAGEVIEWS.value,
                                        'website_id', id, 'website_name', record_infos)
                                tracker_log(
                                    f'Write {pageviews} pageviews to {timeseries_key_name} at {get_ts(parsed_key_name.get("ts"))}',
                                    f'{job_name} - ')
                            else:
                                tracker_log(
                                    f'Write {pageviews} pageviews to {timeseries_key_name} at {get_ts(parsed_key_name.get("ts"))}',
                                    f'{job_name} - warning - ')
                        elif dimension == Dimension.SECTION.value:
                            # record_infos = execute('FT.SEARCH', index_name, f'@id:{{{id}}}', 'RETURN', '5',
                            #                        'name', 'AS', 'section_name', 'website_id', 'website_name')
                            json_key = create_key_name(Type.JSON.value, '', Dimension.SECTION.value, id, '', '')
                            # JSON.GET J::S:46535707faf945b66a11ddf37f8017b2c29bfa3e:: website.id website.name name
                            record_infos = execute('JSON.GET', json_key, 'website.id', 'website.name', 'name')
                            record_infos = json.loads(record_infos)

                            # tracker_log(f'records_infos {json_key} {record_infos}', f'{job_name} - ')
                            # FixMe: HOT FIX -> no records info found
                            if bool(record_infos):
                                execute('TS.ADD', timeseries_key_name, get_ts(parsed_key_name.get("ts")), pageviews,
                                        'ON_DUPLICATE', 'LAST',
                                        'LABELS', 'ts_name', timeseries_name,
                                        'dimension', dimension, Dimension.METRIC.value, Metric.PAGEVIEWS.value,
                                        'section_id', id, 'section_name', record_infos.get('name'),
                                        'website_id', record_infos.get('website.id'), 'website_name',
                                        record_infos.get('website.name'))
                                tracker_log(
                                    f'Write {pageviews} pageviews to {timeseries_key_name} at {get_ts(parsed_key_name.get("ts"))}',
                                    f'{job_name} - ')
                            else:
                                tracker_log(
                                    f'Write {pageviews} pageviews to {timeseries_key_name} at {get_ts(parsed_key_name.get("ts"))}',
                                    f'{job_name} - warning - ')
                        elif dimension == Dimension.PAGE.value:
                            # record_infos = execute('FT.SEARCH', index_name, f'@id:{{{id}}}', 'RETURN', '5',
                            #                        'website_id', 'website_name', 'section_id', 'section_name', 'article_id')
                            json_key = create_key_name(Type.JSON.value, '', Dimension.PAGE.value, id, '', '')
                            # JSON.GET J::P:620070307ea50a61c177f8c3dbb063b708dad92c:: website.id website.name section.id section.name article_id
                            record_infos = execute('JSON.GET', json_key, 'website.id', 'website.name', 'section.id',
                                                   'section.name', 'article_id')
                            record_infos = json.loads(record_infos)

                            # tracker_log(f'records_infos {json_key} {record_infos}', f'{job_name} - ')
                            # FixMe: HOT FIX -> no records info found
                            if bool(record_infos):
                                execute('TS.ADD', timeseries_key_name, get_ts(parsed_key_name.get("ts")), pageviews,
                                        'ON_DUPLICATE', 'LAST',
                                        'LABELS', 'ts_name', timeseries_name,
                                        'dimension', dimension, Dimension.METRIC.value, Metric.PAGEVIEWS.value,
                                        'page_id', id,
                                        'website_id', record_infos.get('website.id'), 'website_name',
                                        record_infos.get('website.name'),
                                        'section_id', record_infos.get('section.id'), 'section_name',
                                        record_infos.get('section.name'),
                                        'article_id', record_infos.get('article_id'))
                                tracker_log(
                                    f'Write {pageviews} pageviews to {timeseries_key_name} at {get_ts(parsed_key_name.get("ts"))}',
                                    f'{job_name} - ')
                            else:
                                tracker_log(
                                    f'Write {pageviews} pageviews to {timeseries_key_name} at {get_ts(parsed_key_name.get("ts"))}',
                                    f'{job_name} - warning - ')

                execute('XADD', output_stream_name, 'MAXLEN', f'{capped_stream}', '*',
                        'ts', parsed_key_name.get('ts'),
                        'status', 'merged',
                        'ids', json.dumps(records.get('ids')))
                # tracker_log(
                #     f'XADD {output_stream_name} MAXLEN {capped_stream} * ts {parsed_key_name.get("ts")} status merged ids {type({json.dumps(records.get("ids"))})} {json.dumps(records.get("ids"))}',
                #     f'{job_name} - ')

                execute('EXPIRE', cms_merge.get('name'), key_expire_duration_sc)

        return records



import sys
import random
import time
import redis
import datetime

r = redis.Redis(host="localhost", port=6379)

s_test_name = 'test'
# s_tracker_name = 'tracker'
s_tracker_name = 'ST:TRACKER::::'
s_articles_name = 's-articles'
s_enrich_name = 's-enrich'
ps_metrics_page = 'pb-page-metrics'

# xadd "tracker" "*" "url" "https://www.rtl.lu/meenung/carte-blanche/a/1874304.html" "uui" "ipdigl0jjm5kwusf5zyg8x6n" "ua" "Mozilla/5.0 (iPhone; CPU iPhone OS 14_4_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0.3 Mobile/15E148 Safari/604.1" "lang" "fr-fr" "dt" "2022-03-09T12:35:01.977Z"
# xadd "tracker" "*" "url" "https://www.rtl.lu/news/international/a/1877030.html" "uui" "ipdigl0ja4pko3l94bm3atkz" "ua" "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.110 Safari/537.36" "lang" "en-US,en;q=0.9" "dt" "2022-03-09T12:35:02.072Z"
# xadd "tracker" "*" "url" "https://www.rtl.lu/mobiliteit/news/a/1877195.html" "uui" "ipdigkwnlnbcwsw0qx1622" "ua" "Mozilla/5.0 (Linux; Android 11; M2007J20CG Build/RKQ1.200826.002; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/98.0.4758.101 Mobile Safari/537.36 [FB_IAB/FB4A;FBAV/356.0.0.28.112;]" "lang" "en-GB,en;q=0.9,fr-LU;q=0.8,fr;q=0.7,de-DE;q=0.6,de;q=0.5,en-US;q=0.4" "dt" "2022-03-09T12:35:02.355Z"
# xadd "tracker" "*" "url" "https://5minutes.rtl.lu/actu/monde/a/1877034.html?didomiConfig.notice.enable=false" "uui" "ipdigkwklyvx9gp53bo5n2o" "ua" "Mozilla/5.0 (Linux; Android 12; SM-A528B Build/SP1A.210812.016; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/98.0.4758.101 Mobile Safari/537.36" "lang" "it-IT,it;q=0.9,fr-FR;q=0.8,fr;q=0.7,de-DE;q=0.6,de;q=0.5,en-GB;q=0.4,en-US;q=0.3,en;q=0.2" "dt" "2022-03-09T12:35:01.880Z"
# xadd "tracker" "*" "url" "https://www.rtl.lu/sport/auto-moto-sport/a/1877315.html" "uui" "ipdigkuzvet7lowu3fgz2de" "ua" "Mozilla/5.0 (iPhone; CPU iPhone OS 15_3_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/15.3 Mobile/15E148 Safari/604.1" "lang" "de-DE,de;q=0.9" "dt" "2022-03-09T12:35:02.505Z"
# xadd "tracker" "*" "url" "https://www.rtl.lu/news/international/a/1877198.html" "uui" "ipdigl08he4ecn19s1nsebbt" "ua" "Mozilla/5.0 (Linux; Android 11; SM-A705FN Build/RP1A.200720.012; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/99.0.4844.58 Mobile Safari/537.36 [FB_IAB/FB4A;FBAV/356.0.0.28.112;]" "lang" "en-GB,en-US;q=0.9,en;q=0.8" "dt" "2022-03-09T12:35:01.978Z"
# xadd "tracker" "*" "url" "https://www.rtl.lu/news/national/a/1876697.html" "uui" "ipdigkv195auv4zc50x7imjo" "ua" "Mozilla/5.0 (SMART-TV; Linux; Tizen 5.5) AppleWebKit/537.36 (KHTML, like Gecko) SamsungBrowser/3.0 Chrome/69.0.3497.106 TV Safari/537.36" "lang" "fr-FR" "dt" "2022-03-09T12:35:03.893Z"
# xadd "tracker" "*" "url" "https://www.rtl.lu/news/national/a/1876686.html?didomiConfig.notice.enable=false" "uui" "null" "ua" "Mozilla/5.0 (iPhone; CPU iPhone OS 15_3_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/15E148" "lang" "de-DE,de;q=0.9" "dt" "2022-03-09T14:32:55.741Z"
# xadd "tracker" "*" "url" "https://www.rtl.lu/news/national/a/1876686.html" "uui" "null" "ua" "Mozilla/5.0 (Linux; Android 11; SM-A705FN Build/RP1A.200720.012; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/99.0.4844.58 Mobile Safari/537.36 [FB_IAB/FB4A;FBAV/356.0.0.28.112;]" "lang" "de-DE,de;q=0.9" "dt" "2022-03-09T14:31:55.741Z"


urls = [
    'https://www.rtl.lu/meenung/carte-blanche/a/1874304.html',
    'https://www.rtl.lu/news/international/a/1877030.html',
    'https://www.rtl.lu/mobiliteit/news/a/1877195.html',
    'https://5minutes.rtl.lu/actu/monde/a/1877034.html',
    'https://www.rtl.lu/news/international/a/1877198.html',
    # 'https://today.rtl.lu/news/luxembourg/a/1963501.html',
    # 'https://www.eldo.lu/aktuell/news/a/119006.html'
]

uuis = [
    'ipdigl0jjm5kwusf5zyg8x6n',
    'ipdigkwnlnbcwsw0qx1622',
    'ipdigkuzvet7lowu3fgz2de',
    'ipdigkv195auv4zc50x7imjo'
]


def create_record():
    # now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    now = round(datetime.datetime.now().timestamp() * 1000)
    return {
        'url': 'https://www.rtl.lu/news/international/a/1877030.html',
        'uui': 'ipdigl0jjm5kwusf5zyg8x6n',
        'dt': now,
        'metadata': {
            'schema_version': 1,
            'schema_name': 'TRACKER',
            'object_name': 'TRACKER'
        }
    }


def create_random_record():
    url = random.choice(urls)
    uui = random.choice(uuis)
    # now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    now = round(datetime.datetime.now().timestamp() * 1000)

    return {
        'url': url,
        'uui': uui,
        'dt': now
    }


def main():
    while True:
        record = create_random_record()
        # record = create_record()
        interval = random.randint(0, 2)
        # r.xadd(s_test_name, record)
        r.xadd(s_tracker_name, record)
        dt = record['dt']
        print(f'{interval} => XADD {s_tracker_name} {record} ({type(dt)})')
        time.sleep(interval)
        # time.sleep(1)


if __name__ == '__main__':
    sys.exit(main())


from rgtracker.common import *
import json
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import ClassVar
from redisgears import executeCommand as execute
from redisgears import log


@dataclass
class Website:
    id: str
    name: str = None
    last_visited: str = None
    last_visited_rounded: field(init=False) = None
    dimension_key: field(init=False) = None
    dimension_ts_key: field(init=False) = None
    metric_pageviews_key: field(init=False) = None
    metric_unique_device_key: field(init=False) = None
    cms_width: ClassVar[int] = 2000
    cms_depth: ClassVar[int] = 5

    def __post_init__(self):
        dt = datetime.fromtimestamp(int(self.last_visited) / 1000) if self.last_visited is not None else None
        rounded_last_visited = int(round_time(dt).timestamp() * 1000)

        self.last_visited_rounded = rounded_last_visited
        self.dimension_key = f'{Type.JSON.value}::{Dimension.WEBSITE.value}:{self.id}::'
        self.dimension_ts_key = f'::{Dimension.WEBSITE.value}:{self.id}:{rounded_last_visited}:'
        self.metric_pageviews_key = f'{Type.CMS.value}::{Dimension.WEBSITE.value}::{rounded_last_visited}:{Metric.PAGEVIEWS.value}'
        self.metric_unique_device_key = f'{Type.HLL.value}::{Dimension.WEBSITE.value}:{self.id}:{rounded_last_visited}:{Metric.UNIQUE_DEVICES.value}'

    def create(self):
        execute('SADD', Dimension.WEBSITE.value, f'{self.id}:{self.name}')
        # log(f'SADD websites {self.name}:{self.id}')
        execute('JSON.SET', self.dimension_key, '.', json.dumps({
            'id': self.id,
            'name': self.name,
            'last_visited': self.last_visited,
            'sections': [],
            'pages': []
        }))
        # log(f'JSON.SET {self.dimension_key}')

    def create_metrics(self):
        try:
            execute('CMS.INITBYDIM', self.metric_pageviews_key, self.cms_width, self.cms_depth)
            # tracker_log(f'CMS.INITBYDIM {self.metric_pageviews_key}', f'Website - create_metrics - ')
        except Exception as e:
            # tracker_log(f'{e} - {self.metric_pageviews_key}', 'Website - ')
            pass

        try:
            execute('PFADD', self.metric_unique_device_key)
            # tracker_log(f'PFADD {self.metric_unique_device_key}', f'Website - create_metrics - ')
        except Exception as e:
            # tracker_log(f'{e} - {self.metric_pageviews_key}', 'Website - ')
            pass

    def incr_metrics(self, device_id):
        execute('CMS.INCRBY', self.metric_pageviews_key, self.id, 1)
        # tracker_log(f'CMS.INCRBY {self.metric_pageviews_key} {self.id} 1', f'Website - incr_metrics - ')
        execute('PFADD', self.metric_unique_device_key, device_id)
        # tracker_log(f'PFADD {self.metric_unique_device_key} {device_id}', f'Website - incr_metrics - ')

    def expire_new_metrics(self):
        ttl_pageviews = execute('TTL', self.metric_pageviews_key)
        # tracker_log(f'TTL {self.metric_pageviews_key} {ttl_pageviews}', 'Website - expire new - ')
        if ttl_pageviews == -1:
            expire_pg = execute('EXPIRE', self.metric_pageviews_key, 3600)
            # tracker_log(f'Expire key {self.metric_pageviews_key} = {expire_pg}', f'Website - expire_new_metrics - ')

        ttl_unique_devices = execute('TTL', self.metric_unique_device_key)
        # tracker_log(f'TTL {self.metric_unique_device_key} {ttl_unique_devices}', 'Website - expire new ')
        if ttl_unique_devices == -1:
            expire_ud = execute('EXPIRE', self.metric_unique_device_key, 3600)
            # tracker_log(f'Expire key {self.metric_unique_device_key} = {expire_ud}', f'Website - expire_new_metrics - ')

    # Todo: put expire time in method params
    def expire_metrics(self):
        # Todo: compare current time with self.last_visited and apply : (self.last_visited - current_time) + expire_time
        # That leaves time for further merger services

        expire_pg = execute('EXPIRE', self.metric_pageviews_key, 3600)
        # tracker_log(f'EXPIRE {self.metric_pageviews_key} = {expire_pg}', f'Website - expire_metrics - ')

        expire_ud = execute('EXPIRE', self.metric_unique_device_key, 3600)
        # tracker_log(f'EXPIRE {self.metric_unique_device_key} = {expire_ud}', f'Website - expire_metrics - ')

    def expire_record_metrics(self):
        ttl_unique_devices = execute('TTL', self.metric_unique_device_key)
        # tracker_log(f'TTL {self.metric_unique_device_key} {ttl_unique_devices}', 'Website - expire_record_metrics - ')
        if ttl_unique_devices == -1:
            expire_ud = execute('EXPIRE', self.metric_unique_device_key, 3600)
            # tracker_log(f'Expire key {self.metric_unique_device_key} = {expire_ud}', f'Website - expire_record_metrics - ')

    def is_exists(self):
        x = execute('EXISTS', self.dimension_key)
        # log(f'EXISTS {self.dimension_key} => {x}')
        return x

    def has_metrics(self):
        # check for weird timestamp
        x = execute('EXISTS', self.metric_pageviews_key)
        # log(f'EXISTS {self.metric_pageviews_key} => {x}')
        # z = execute('EXISTS', self.metric_unique_device_key)
        # log(f'EXISTS {self.metric_unique_device_key} => {z}')
        # return x + z
        return x

    def update_last_visited(self):
        execute('JSON.SET', self.dimension_key, '.last_visited', self.last_visited)
        # log(f'update_last_visited: JSON.SET {self.dimension_key} .last_visited {self.last_visited}')

    def notification_of_new_website(self, stream_names):
        # tracker_log(f'{self.last_visited_rounded} new website {self.id}', f'Website - notification_of_new_website - ')
        execute('XADD', stream_names.get('website_pageviews'), 'MAXLEN', '8640', '*',
                'ts', self.last_visited_rounded,
                'status', 'new',
                'ids', self.id)
        execute('XADD', stream_names.get('website_unique_devices'), 'MAXLEN', '8640', '*',
                'ts', self.last_visited_rounded,
                'status', 'new',
                'ids', self.id)

    def notification_of_active_websites(self, stream_names):
        prev_unix_ts = int(
            (datetime.fromtimestamp(int(self.last_visited_rounded) / 1000) - timedelta(minutes=1)).timestamp() * 1000)

        # Todo: create index_name with method instead of hard coding it
        ids_numbers = int(
            execute('FT.SEARCH', 'I::W:::', f'@last_visited:[{prev_unix_ts} {self.last_visited_rounded}]', 'LIMIT', 0,
                    0)[0])
        ids_raw = execute('FT.SEARCH', 'I::W:::', f'@last_visited:[{prev_unix_ts} {self.last_visited_rounded}]',
                          'LIMIT', 0, ids_numbers, 'RETURN', '1', 'id')
        ids = []
        for d in ids_raw:
            if type(d) is list:
                ids.append(d[-1])

        # FixMe: HOT FIX
        if len(ids) > 0:
            tracker_log(
                f'[{datetime.fromtimestamp(prev_unix_ts / 1000)} - {datetime.fromtimestamp(int(self.last_visited_rounded) / 1000)}] {len(ids)} active websites',
                f'Website - notification_of_active_websites - ')
            execute('XADD', stream_names.get('website_pageviews'), 'MAXLEN', '8640', '*',
                    'ts', self.last_visited_rounded,
                    'status', 'active',
                    'ids', json.dumps(ids))
            execute('XADD', stream_names.get('website_unique_devices'), 'MAXLEN', '8640', '*',
                    'ts', self.last_visited_rounded,
                    'status', 'active',
                    'ids', json.dumps(ids))
        else:
            tracker_log(
                f'[{datetime.fromtimestamp(prev_unix_ts / 1000)} - {datetime.fromtimestamp(int(self.last_visited_rounded) / 1000)}] {len(ids)} active websites',
                f'Website - notification_of_active_websites - warning - ')

    def send_to_enrich(self):
        execute('XADD', 'ST:ENRICH:W:::', 'MAXLEN', '10', '*', 'key', self.dimension_key)


def load_website(website, section, page, device, output_stream_names):
    if website.is_exists() != 1:
        website.create()
        website.create_metrics()
        website.incr_metrics(device.id)
        website.expire_new_metrics()
        website.notification_of_new_website(stream_names=output_stream_names)
        website.send_to_enrich()
    else:
        if website.has_metrics() < 1:
            website.create_metrics()
            website.incr_metrics(device.id)
            website.expire_metrics()
            website.update_last_visited()
            website.notification_of_active_websites(stream_names=output_stream_names)
        else:
            website.incr_metrics(device.id)
            website.expire_record_metrics()
            website.update_last_visited()



from rgtracker.record import *
from rgtracker.tracker import *
from rgtracker.common import *
from rgtracker.website import *
from rgtracker.section import *
from rgtracker.page import *
from rgtracker.device import *
from rgtracker.pageviews import *

# from redis import Redis
#
# redis_dev = Redis(host="10.61.228.218", port=6379, password='PiPaPo-2021.12')
# redis_local = Redis(host="localhost", port=6379)
#
# player_data = redis_dev.xrange('ST:PLAYER::::', 1663767601740, 1663767808901)
#
# for pd in player_data:
#     redis_local.xadd('ST:PLAYER::::', pd[1])
#     print(f'XADD ST:PLAYER:::: {pd[1]}')

# foreach(lambda record: log(f'{record}')). \
# foreach(lambda record: execute('TOPK.ADD', 'topk', {record.value.url})). \
GB("StreamReader"). \
    foreach(lambda record: execute('TOPK.ADD', 'topk', {record.get('value').get('url')})). \
    run('ST:PLAYER::::', trimStream=False)
# run('ST:PLAYER::::', trimStream=False, fromId="1663766604999")

# gears-cli run --host localhost --port 6379 src/jobs/sandbox/sandbox-player.py
